"""Routes for Backends in CMB.

This module contains REST API function calls which permits backends to poll for,
execute, and store results of code-profilings.

The addition/removal of backends is done through the admin interface.
The REST API within only permit the GETting of new submissions/runs by a
backend, the POSTing of their results back to the server, in addition to
refreshing the server so that it may know when a backend has stopped polling.
"""

import os
from os import path as op
from multiprocessing import Lock
from datetime import datetime, timedelta

from scp import SCPClient
from sqlalchemy import desc
from werkzeug import secure_filename
from flask import Blueprint, jsonify, request, Response

from cmb_utils.wrappers import myWrappers
from database.models import db, Run, Backend
from cmb_utils.network_utils import connectSSHClient
from cmb_utils.backend_utils import ProfilingStepReady
# from cmb_utils.backend_utils import is_file_content_text
from cmb_utils import getCurrentTimeString, execute_os_command

atomic_db_access_lock = Lock()  # Used to ensure atomic retrieval of potential runs
backend_routes = Blueprint('backend_routes', __name__, None)


def check_for_available_runs_by_backend(backend, LIFO_order=False):
    # Check first if a run is already assigned to backend
    already_assigned_run = Run.query.filter(
        Run.backend_assigned == backend,
        Run.dequeued.isnot(None),
        Run.finished.is_(None)).first()

    if already_assigned_run is not None:
        return already_assigned_run

    # If not, see if there's a new one to assign
    untaken_runs = []
    atomic_db_access_lock.acquire()     # Acquire mutex lock
    if LIFO_order:
        untaken_runs = Run.query.\
            filter_by(backend_assigned=None).\
            order_by(desc(Run.enqueued)).all()
    else:
        untaken_runs = Run.query.\
            filter_by(backend_assigned=None).\
            order_by(Run.enqueued).all()

    for run in untaken_runs:
        if (run.submission.problem.architecture == backend.architecture and
           run.submission.problem.software_required <= set(backend.softwares)):
            # Update run that it's going to get processed
            run.backend_assigned = backend
            run.dequeued = datetime.now()
            db.session.add(run)
            db.session.commit()
            atomic_db_access_lock.release()     # Release mutex lock
            return run

    # If no run was found in for-loop:
    atomic_db_access_lock.release()     # Release mutex lock
    return None


@backend_routes.route('/api/backends/poll/', methods=['GET'])
@myWrappers.backends_auth_required
def backend_query_for_submission_runs():
    # Get ID of the backend for this polling
    # Force python/flask to read the request's json data with 'correct' mimetype
    json_data = request.get_json(force=True)

    # Confirm that this is an existing backend
    backend_id = json_data.get('backend_id')
    backend = Backend.query.filter_by(id=backend_id).first()
    if backend is None:
        return Response(
            "No backend found with id={}.\n"
            "You must supply an integer corresponding to the database"
            " ID of an existing backend.".format(backend_id), 400)

    # Register that the backend has queried for submission.
    backend.last_query = datetime.now()
    db.session.add(backend)
    db.session.commit()

    # Get the new potential run
    potential_run = None
    try:
        potential_run = check_for_available_runs_by_backend(backend)
    except Exception as e:
        print(getCurrentTimeString() + "Following Exception occured during "
              "fetching of a new/next run for backend #{}\n\n{}\n".format(
              backend_id, e))
        atomic_db_access_lock.release()
    if potential_run is None:
        # FUTURE WORK:
        # Perhaps wait an interval, and check for new request,
        # instead of forcing backend to poll again.
        return Response("No available runs found in DB for this backend.", 204)

    # Create response object w/data
    return_json = dict(filepath=potential_run.compiled_files_path, run_id=potential_run.id)
    response = jsonify(message="Run found.", data=return_json)
    response.status_code = 200
    return response


@backend_routes.route('/api/backends/update_results/', methods=['POST'])
@myWrappers.backends_auth_required
def backend_update_run_values():
    json_data = request.get_json(force=True)

    # Confirm that this is an existing backend
    backend_id = json_data.get('backend_id')
    backend = Backend.query.filter_by(id=backend_id).first()
    if backend is None:
        return Response(
            "No backend found with id={}.\n"
            "You must supply an integer corresponding to the database"
            " ID of an existing backend.".format(backend_id), 400)

    # Confirm that the backend is reporting on an existing, viable, Run
    run_id = json_data.get('run_id')
    current_run = Run.query.filter_by(id=run_id).first()
    if current_run is None:
        return Response(
            "No Run found with id={}.\n"
            "You must supply an integer corresponding to the database"
            " ID of an existing backend.".format(backend_id), 400)

    backend.last_query = current_run.finished = datetime.now()
    submission_of_run = current_run.submission
    db.session.add(current_run)
    db.session.add(backend)
    db.session.commit()

    print(getCurrentTimeString() + "Received completed backend profiling execution:\n".format(run_id))
    print(json_data)
    print("")

    # Deal with the case that it didn't finish successfully on backend
    if not json_data['was_profiling_successfully_completed?']:
        submission_of_run.state = "failed"
        current_run.msg = json_data['msg']

        check_step = ProfilingStepReady(0)
        # While-loop for future readability, even though
        # it would've been "faster" to just create the
        # ProfilingStepReady corresponding to the value of
        # json_data['last_successful_step'] instead...
        while check_step < json_data['last_successful_step']:
            check_step += 1

        failed_step = check_step + 1
        print(getCurrentTimeString() + "Profiling failed during step {} on backend.".format(failed_step))
        # If the failed step is during checking correctness of small tests,
        if failed_step > 4:
            submission_of_run.msg = (
                "Profiling of uploaded code failed on "
                "backend between '{}' and '{}'".format(
                    check_step.status, failed_step.status))
        else:
            submission_of_run.msg = current_run.msg

        # Save and report
        db.session.add(backend)
        db.session.add(current_run)
        db.session.add(submission_of_run)
        db.session.commit()
        response = jsonify(message="Ack")
        response.status_code = 200
        return response

    # Copy over tmpfile with big test(s)'s output from backend,
    # and save file in local folder cmb-flask/problems/<problem_name>/problemIO/
    remote_path = json_data['output_file_path']
    local_path = secure_filename(current_run.submission.problem.name).lower()
    local_path += "/problemIO/{}_solution.txt".format(run_id)
    local_path = op.join(op.expandvars(os.environ['UPLOAD_FOLDER']), local_path)
    try:
        cmb_ssh_client = connectSSHClient(
            hostname=backend.ip_address, port=22,
            username=os.environ['CMB_USER'], private_key_pw=os.environ['CMB_SERVER_SSH_PW'])
        print(getCurrentTimeString() + "\tMade SSH connection client.")
        try:
            # Copy over files from server
            print(getCurrentTimeString() + "\tOpening SCP connection with SSH client...")
            with SCPClient(cmb_ssh_client.get_transport()) as scp:
                print(getCurrentTimeString() + "\tCopying files...")
                scp.get(remote_path, local_path)
            print(getCurrentTimeString() + "\tClosing SCP connection client.")

            print(getCurrentTimeString() + "\tDeleting tempfile on backend...")
            rm_cmd = "rm {}".format(remote_path)
            inpt, outpt, err = cmb_ssh_client.exec_command(rm_cmd)
            output_str, error_str = outpt.read().decode(), err.read().decode()
            if output_str != "" and error_str != "":
                print(getCurrentTimeString() + "\tDeletion of tempfile {} on backend {} failed!".format(
                    remote_path, backend_id))
                print("\t\t" + rm_cmd)
                print("\n\tstdout: " + output_str)
                print("\n\tstderr: " + error_str)
            else:
                print(getCurrentTimeString() + "\tTempfile {} on backend {} deleted!".format(
                    remote_path, backend_id))
        except Exception as e:
            error_message = "\tFollowing Exception occured during SCP operations to IP:{}:\n".format(backend.ip_address)
            error_message += "\t{}".format(e)
            print(getCurrentTimeString() + error_message)
    except Exception as e:
        error_message = "\tFollowing Exception occured during SSH/SCP Operations:\n"
        error_message += "\t{}".format(e)
        print(getCurrentTimeString() + error_message)
    finally:
        print(getCurrentTimeString() + "\tClosing SSH connection client.")
        cmb_ssh_client.close()

        if not op.exists(local_path):
            # Should never happen...
            print("\n\tFILE '{}' WAS NOT SUCCESSFULLY SCP'ED FROM BACKEND:\n\t\t{}\n".format(remote_path, backend))

    # Run checker executable (which should be located in directory to which the
    # tempfile was copied).
    os.chdir(op.dirname(local_path))
    execute_chckr_cmd = ["./checker", "input.txt", local_path, "answer.txt"]
    execute_chckr_dict = execute_os_command(cmd=execute_chckr_cmd, timeout=None)

    # Parse checker output, see if output passed it or not
    submission_of_run.state = "finished"
    chckr_output = execute_chckr_dict['stdout'].split("\n")
    if (execute_chckr_dict['returncode'] != 0 and str(chckr_output[0]).lower() != "ok"):
        print(getCurrentTimeString() + "Profiling big test(s) output failed correctness test.")
        submission_of_run.msg = "Failed big correctness test"
        submission_of_run.state = "failed"

        # Save and report
        backend.last_query = datetime.now()
        db.session.add(backend)
        db.session.add(current_run)
        db.session.add(submission_of_run)
        db.session.commit()
        response = jsonify(message="Ack")
        response.status_code = 200
        return response

    os.remove(local_path)

    # set correctness:
    # These checks are complicated to work with,
    # it should instead be decided that the string
    # "ok" should always be first (or last) line of output from checker.cpp
    # but not both!!!
    if (len(chckr_output) == 1 and
       str(chckr_output[0]).lower() != "ok"):
        current_run.goodness_value = chckr_output[0]
    elif (len(chckr_output) >= 2 and
          str(chckr_output[0]) and
          str(chckr_output[0]).lower() != "ok"):
        current_run.goodness_value = chckr_output[1]
    else:
        current_run.goodness_value = None

    # set energy and timing values
    current_run.power = json_data['power']
    submission_of_run.time = current_run.runtime = json_data['time']
    submission_of_run.energy = current_run.energy = json_data['energy']
    submission_of_run.edp = current_run.edp = json_data['time'] * json_data['energy']

    # Finish setting book keeping values
    submission_of_run.goodness_value = current_run.goodness_value
    submission_of_run.msg = None

    # Save and report
    backend.last_query = datetime.now()
    db.session.add(backend)
    db.session.add(current_run)
    db.session.add(submission_of_run)
    db.session.commit()
    response = jsonify(message="Ack")
    response.status_code = 200
    return response


@backend_routes.route('/api/backends/status/', methods=['GET'])
@backend_routes.route('/api/backends/status/<int:optional_time_limit>/', methods=['GET'])
@myWrappers.sameIP_required
def check_last_query_of_backends(optional_time_limit=20):
    backends = Backend.query.all()

    late_backends = [("backend.id", "backend.last_query")]
    should_have_polled_since = datetime.now()
    should_have_polled_since -= timedelta(seconds=optional_time_limit)
    for backend in backends:
        # Late_backend = namedtuple('backend_id', 'last_query')
        if (backend.last_query is None or
           backend.last_query < should_have_polled_since):
            # late_backends += Late_backend(backend.id, backend.last_query)
            late_backends += [(backend.id, backend.last_query)]

    # If one or more backends are late:
    if len(late_backends) > 1:
        busy_backends = []
        for backend in late_backends:
            runs_assigned_to_backend = Run.query.filter(
                Run.backend_assigned == backend,
                Run.dequeued.isnot(None),
                Run.finished.is_(None)).first()
            if runs_assigned_to_backend:
                busy_backends += backend.id

        return jsonify(
            returncode=1, late_backends=late_backends,
            time_limit="{} second(s).".format(optional_time_limit),
            current_time_and_date="Current time and date on server: "
            "{}".format(getCurrentTimeString()),
            backends_currently_busy_with_a_code_profiling=busy_backends,
            message="Some backends have not polled within the given time limit.")

    # If no backends were late:
    return jsonify(
        returncode=0,
        time_limit="{} second(s).".format(optional_time_limit),
        current_time_and_date="Current time and date on server: "
        "{}".format(getCurrentTimeString()),
        message="All backends have polled within the given time limit.")
