from time import sleep
from socket import gethostname
from os import path as op, environ as envvars

import requests
import simplejson as json
from requests.exceptions import ConnectionError

from odroid_xu3 import cleanup
from cmb_utils import getCurrentTimeString
from cmb_utils.backend_utils import ProfilingStepReady
from odroid_xu3.odroid_xu3_specific_implementation import prepare_backend_run, compile_submission
from odroid_xu3.odroid_xu3_specific_implementation import execute_small_tests, execute_checker_small_tests
from odroid_xu3.odroid_xu3_specific_implementation import prepare_profiling, execute_big_tests
from odroid_xu3.odroid_xu3_specific_implementation import parse_and_prepare_profiling_data

# Mandatory and required environment variables
CMB_USER = op.expandvars(envvars['CMB_USER'])
CMB_SERVER = op.expandvars(envvars['CMB_SERVER'])
CMB_SERVER_PORT = op.expandvars(envvars['CMB_SERVER_PORT'])
POLLING_URL = "http://" + CMB_SERVER + ":" + CMB_SERVER_PORT + "/api/backends/poll/"
RESULTS_URL = "http://" + CMB_SERVER + ":" + CMB_SERVER_PORT + "/api/backends/update_results/"

# Non-mandatory environment variables
SLEEP_PERIOD, SMPLNG_INTRVL = envvars.get('SLEEP_PERIOD'), envvars.get('samplingInterval')
SLEEP_PERIOD = SLEEP_PERIOD if isinstance(SLEEP_PERIOD, int) else 12
SMPLNG_INTRVL = SMPLNG_INTRVL if isinstance(SMPLNG_INTRVL, int) else 1000


def poll_request(timeout_limit_in_seconds=4.00):
    print(getCurrentTimeString() + "Requesting next code submission from server...")
    status_code, request = 503, None
    try:
        # Wait until backend gets a code submission to profile
        req = requests.get(
            POLLING_URL, timeout=timeout_limit_in_seconds,
            auth=(gethostname(), envvars['BACKEND_SHARED_PW']),
            data=json.dumps(dict(backend_id=envvars['BACKEND_ID'])))
        status_code, request = req.status_code, req
    except ConnectionError:
        print(getCurrentTimeString() + "Unable to reach {}.".format(POLLING_URL))
    except requests.exceptions.Timeout:
        print(getCurrentTimeString() + "GET request to {} timed out with "
              "timeout of {} seconds.".format(POLLING_URL, timeout_limit_in_seconds))

    if status_code == 200:
        print("\n" + getCurrentTimeString() + "Code submission received from server.")
    else:
        print(getCurrentTimeString() + "No code submission received from server.")

    return status_code, request


def push_results(results_json):
    while True:
        try:
            print(getCurrentTimeString() + "POST'ing profile execution data back to server:\n")
            print(results_json)
            print("")
            req = requests.post(
                RESULTS_URL, timeout=300.00,
                data=json.dumps(results_json, indent=2),
                headers={'Content-Type': 'application/json'},
                auth=(gethostname(), envvars['BACKEND_SHARED_PW']))
            status_code, request = req.status_code, req
        except ConnectionError:
            print(getCurrentTimeString() + "Unable to reach {}.".format(RESULTS_URL))
            sleep(SLEEP_PERIOD / 2)
            continue

        if status_code == 200:
            break

    return status_code, request


def main():
    while True:     # "main" infinite loop
        status_code, polling_request = poll_request()
        while status_code != 200:
            sleep(SLEEP_PERIOD)
            status_code, polling_request = poll_request()

        print(getCurrentTimeString() + "Parsing code submission received from server...")
        try:
            request_json_data = polling_request.json()
        except Exception as e:
            print(getCurrentTimeString() +
                  "Following Exception occured during parsing of received request's JSON buffer:\n\n{}\n"
                  "\tRe-fetching code submission from server.\n".format(e))
            continue
        print(getCurrentTimeString() + "Code submission received from server parsed:")
        print(request_json_data)

        try:
            print(getCurrentTimeString() + "Preparing backend code profiling run...")
            profile_dict = run_backend_profiling_run(request_json_data, sleep_period=SLEEP_PERIOD)
        except Exception as e:
            print(getCurrentTimeString() +
                  "Following Exception occured during execution of backend profiling run:\n\n{}\n"
                  "\tRe-fetching code submission from server.\n".format(e))
            continue
        finally:
            cleanup(request_json_data['data']['run_id'])

        # Return data
        push_results(profile_dict)


def run_backend_profiling_run(request_json, sleep_period, max_retries=10):
    data = request_json['data']
    run_id, server_filepath = data['run_id'], data['filepath']

    # Folder depth necessary due to the
    # (legacy) Makefile used for C/C++ code submissions
    destination_folder = "~/cmb-board/mountBlanc/files/" + str(run_id) + "/"
    destination_folder = op.expanduser(destination_folder)

    # SCP/Fetch relevant/needed stuff for this code submission from server
    counter = 1
    profile_dict = dict(
        run_id=run_id,
        backend_id=envvars['BACKEND_ID'],
        last_successful_step=ProfilingStepReady(0))
    profile_dict.update(
        prepare_backend_run(
            destination_folder=destination_folder,
            server_filepath=server_filepath, sleep_period=sleep_period))
    while (profile_dict['was_last_operation_a_success?'] is False and
           counter < max_retries):
        error_msg = "Re-trying the fetching of files from server, "
        error_msg += "\tand the preparation of these on the backend, "
        error_msg += "for the {} time.".format(counter)
        print(getCurrentTimeString() + error_msg)
        profile_dict = prepare_backend_run(
            destination_folder=destination_folder,
            server_filepath=server_filepath, sleep_period=sleep_period)
        counter += 1
    if profile_dict['was_last_operation_a_success?'] is not True:
        profile_dict['was_profiling_successfully_completed?'] = False
        cleanup(run_id)
        return profile_dict
    profile_dict['last_successful_step'] += 1

    # Attempt to compile the submission, abort if it fails
    profile_dict['work_folder'] = destination_folder
    profile_dict = compile_submission(profile_dict=profile_dict)
    if profile_dict['was_last_operation_a_success?'] is not True:
        profile_dict['was_profiling_successfully_completed?'] = False
        return profile_dict
    profile_dict['last_successful_step'] += 1

    # Run the "small test(s)" with the compiled code submission
    profile_dict = execute_small_tests(profile_dict=profile_dict)
    if profile_dict['was_last_operation_a_success?'] is not True:
        profile_dict['was_profiling_successfully_completed?'] = False
        return profile_dict
    profile_dict['last_successful_step'] += 1

    # Compile and execute the checker for the C/C++ code submission
    profile_dict = execute_checker_small_tests(profile_dict=profile_dict)
    if profile_dict['was_last_operation_a_success?'] is not True:
        profile_dict['was_profiling_successfully_completed?'] = False
        return profile_dict
    profile_dict['last_successful_step'] += 1

    # Prepare for the actual profiling run,
    # warming up board, starting energy profiling tool(s), etc.
    profile_dict = prepare_profiling(profile_dict=profile_dict, sampling_interval=SMPLNG_INTRVL)
    if profile_dict['was_last_operation_a_success?'] is not True:
        profile_dict['was_profiling_successfully_completed?'] = False
        return profile_dict
    profile_dict['last_successful_step'] += 1

    # Execute the actual profiling run measuring the
    # energy and time consumption of code submission.
    profile_dict = execute_big_tests(profile_dict=profile_dict)
    if profile_dict['was_last_operation_a_success?'] is not True:
        profile_dict['was_profiling_successfully_completed?'] = False
        return profile_dict
    profile_dict['last_successful_step'] += 1

    profile_dict = parse_and_prepare_profiling_data(profile_dict=profile_dict)
    if profile_dict['was_last_operation_a_success?'] is not True:
        profile_dict['was_profiling_successfully_completed?'] = False
        return profile_dict
    profile_dict['last_successful_step'] += 1

    profile_dict['was_profiling_successfully_completed?'] = True
    return profile_dict


if __name__ == "__main__":
    main()
